`timescale 1ns / 1ns
//////////////////////////////////////////////////////////////////////////////////
// Module Name:    uart_tx 
// Target Devices: Spartan 6 FPGA
// Tool versions: ISE WebPACK 14.1
// Description: This module implements a UART transmitter with configurable parameters.


//////////////////////////////////////////////////////////////////////////////////
module uart_tx( reset, clk16x, thr, thr_wr_strobe, databits, stopbits, parity_en, parity_even, parity_stick, sout, THRE, TXME );

input  wire	reset;			 // global reset
input  wire clk16x;			 // transmitter clock,16 times buad rate
input  wire [7:0] thr;		 // transmitter hold register
input  wire thr_wr_strobe;	 // thr write strobe signal
input  wire [1:0] databits; // select the number of data bits
input  wire [1:0] stopbits; // select the number of stop bits
input  wire parity_en;		 // enable parity check
input  wire parity_even;	 // '0'--odd parity,'1'--even parity
input  wire parity_stick;
output wire sout;				 // serial output,or named as tx line
output wire THRE;		 		 // status output,indicating transmitter holding register is empty;
output wire TXME;		 		 // status output,indicating transmitter holding register and shift register are both empty;

//state definition
parameter	start	       =  3'b000;
parameter	shift	       =  3'b001;
parameter	parity	    =  3'b010;
parameter	stop_1bit	 =  3'b011;
parameter	stop_halfbit =  3'b100;
parameter	stop_2bit	 =  3'b101;
parameter	start1       =  3'b110;


reg [7:0] tsr;	// transimitter shift register
reg [2:0] tx_state;	//state register
reg [2:0] tx_cnt;
reg [3:0] count;

reg tx_output; // serial data output
reg tx_parity; //
reg thr_empty; // transmitter holding register is empty
reg tsr_empty; // transmitter shift register is empty

// these registers indicate the current state of FSM
reg tx_in_start_state;
reg tx_in_shift_state;
reg tx_in_stop_state;

// 1 clk16x clock delayed state of tx_in_start_state,tx_in_shift_state,and tx_in_stop_state
reg tx_in_start_state1;
reg tx_in_shift_state1;
reg tx_in_stop_state1;

// transmitter clock enable singals
reg tx_clken_a;
reg tx_clken_b;




always@( posedge clk16x, posedge reset )
begin
	if( reset == 1'b1 )
	begin
		tx_cnt <= 3'b000;
		tsr <= 8'b0000_0000;
		tx_output <= 1'b1;
		tx_parity <= 1'b1;
		tx_state <= start;
	end
	else
	begin
		case( tx_state )
			start:
				begin
					if( ( thr_empty == 1'b0 )&&( tx_clken_a == 1'b1 ) )
						begin
							tsr <= thr;
							tx_parity <= ~parity_even; // set it when odd parity is selected( parityeven = 0 ),clear it when even parity is selected( parityeven = 1 )
							tx_output <= 1'b0; // start bit
							tx_cnt <= 3'b000;
							tx_state <= shift;
						end
					else
						begin
							tx_output <= 1'b1;
						end
				end
				
			shift:
				begin
					if( tx_clken_a == 1'b1 )
					begin	// shift serial data out
						tsr <= { 1'b0,tsr[7:1] };
						tx_parity <= tx_parity ^ tsr[0];
						tx_output <= tsr[0];
						tx_cnt <= tx_cnt + 1'b1;
						if( ( databits == 2'b00 && tx_cnt == 4 ) || // 5-bit data
						    ( databits == 2'b01 && tx_cnt == 5 ) || // 6-bit data
						    ( databits == 2'b10 && tx_cnt == 6 ) || // 7-bit data
						    ( databits == 2'b11 && tx_cnt == 7 ) 	  // 8-bit data
						  )
						  begin
							if( parity_en == 1'b0 )
								begin
									tx_state <= stop_1bit;
								end
							else
								begin
									tx_state <= parity;
								end
						  end
					end
				end
				
			parity:
				begin
					if( tx_clken_a == 1'b1 )
						begin
							if( parity_stick == 1'b0 )
								begin
									tx_output <= tx_parity;	// output parity bit
								end
							else
								begin
									tx_output <= ~parity_even;	// forced parity bit
								end
							tx_state <= stop_1bit;	
						end		
				end
				
			stop_1bit:
				begin
					if( tx_clken_a == 1'b1 )
						begin
							tx_output <= 1'b1;	
							if( stopbits == 2'b00 )
								tx_state <= start;
							else if( stopbits == 2'b01 )
								tx_state <= stop_halfbit;
							else
								tx_state <= stop_2bit;
						end				
				end
			
			stop_halfbit:
				begin
					if( tx_clken_b == 1'b1 )
						begin
							tx_output <= 1'b1;
							tx_state  <= start;
						end
				end
				
			stop_2bit:
				begin
					if( tx_clken_a == 1'b1 )
						begin
							tx_output <= 1'b1;
							tx_state <= start;
						end
				end
				
			default:
				tx_state <= start;
				
		endcase
		
	end
							
end
	
always@( posedge clk16x, posedge reset )
begin
	if( reset == 1'b1 )
		tsr_empty <= 1'b1;
	else
		begin
			if( ( tx_in_stop_state == 1'b1 )&&( tx_in_stop_state1 == 1'b0 ) )
				tsr_empty <= 1'b1;
			else if( ( tx_in_shift_state == 1'b1 )&&( tx_in_shift_state1 == 1'b0 ) )
				tsr_empty <= 1'b0;
		end
end		
		
		
always@( posedge clk16x, posedge reset )
begin
	if( reset == 1'b1 )
		thr_empty <= 1'b1;
	else
		begin
			if( thr_wr_strobe == 1'b1 )
				thr_empty <= 1'b0;
			else if( ( tx_in_shift_state == 1'b1 )&&( tx_in_shift_state1 == 1'b0 ) ) // set thr_empty flag to 1 when data is transferred from thr to tsr
				thr_empty <= 1'b1;
		end
end


//delayed signal for edge detection
always@( posedge clk16x, posedge reset )
begin
	if( reset == 1'b1 )
		begin
			tx_in_start_state1 <= 1'b1;
			tx_in_shift_state1 <= 1'b1;
			tx_in_stop_state1  <= 1'b1;
		end
	else
		begin
			tx_in_start_state1 <= tx_in_start_state;
			tx_in_shift_state1 <= tx_in_shift_state;
			tx_in_stop_state1  <= tx_in_stop_state;			
		end
end


//transmitter FSM state indication signals
always@( posedge clk16x, posedge reset )
begin
	if( reset == 1'b1 )
		tx_in_start_state <= 1'b0;
	else
		begin
			if( tx_state == start )
				tx_in_start_state <= 1'b1;
			else
				tx_in_start_state <= 1'b0;
		end		
end


always@( posedge clk16x, posedge reset )
begin
	if( reset == 1'b1 )
		tx_in_shift_state <= 1'b0;
	else
		begin
			if( tx_state == shift )
				tx_in_shift_state <= 1'b1;
			else
				tx_in_shift_state <= 1'b0;
		end		
end


always@( posedge clk16x, posedge reset )
begin
	if( reset == 1'b1 )
		tx_in_stop_state <= 1'b0;
	else
		begin
			if( tx_state == stop_1bit )
				tx_in_stop_state <= 1'b1;
			else
				tx_in_stop_state <= 1'b0;
		end		
end

//generate tx_clken_a/tx_clken_b signals
always@( posedge clk16x, posedge reset )
begin
	if( reset == 1'b1 )
		begin
			count <= 4'b1111;
			tx_clken_a <= 1'b0;
			tx_clken_b <= 1'b0;
		end
	else
		begin
			if( count == 4'b0000 ) 
				tx_clken_a <= 1'b1;
			else
				tx_clken_a <= 1'b0;
				
			if( count == 4'b1000 )
				tx_clken_b <= 1'b1;
			else
				tx_clken_b <= 1'b0;
				
			if( ( tx_in_start_state == 1'b1 )&&( tx_in_start_state1 == 1'b0 ) )
				count <= 4'b0011;
			else
				count <= count + 1'b1;
		end
end

assign THRE = thr_empty;
assign TXME = ( ( thr_empty == 1'b1 )&&( tsr_empty == 1'b1 ) ) ? 1'b1 : 1'b0;
assign sout =  tx_output;

endmodule
